(*
	LablScintilla - Scintilla [LablGTK] binding for Objective Caml

	author: Robert Roessler
	email: robertr@rftp.com
	web: http://www.rftp.com

--------------------------------------------------------------------------------
Copyright (c) 2005-2007 Robert Roessler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Robert Roessler nor the names of its contributors
      may be used to endorse or promote products derived from this
      software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
--------------------------------------------------------------------------------
*)

(* scintilla.ml - implementation of Scintilla [LablGTK] binding for OCaml *)

open Gaux
open GObj
open Gtk

type scintilla = [container | `scintilla]

(* mapping of "core" Scintilla types *)
type opaque_ptr = nativeint (* MUST hold a C ptr; 64-bit examination required *)
type document = opaque_ptr
type text_range = {cpMin: int; cpMax: int}

type cache_mode =
	| SC_CACHE_NONE | SC_CACHE_CARET | SC_CACHE_PAGE | SC_CACHE_DOCUMENT

type caret_policy = CARET_SLOP | CARET_STRICT | CARET_JUMPS | CARET_EVEN

type case_mode = SC_CASE_MIXED | SC_CASE_UPPER | SC_CASE_LOWER

type char_set =
	| SC_CHARSET_8859_15 | SC_CHARSET_ANSI | SC_CHARSET_ARABIC | SC_CHARSET_BALTIC
	| SC_CHARSET_CHINESEBIG5 | SC_CHARSET_CYRILLIC | SC_CHARSET_DEFAULT
	| SC_CHARSET_EASTEUROPE | SC_CHARSET_GB2312 | SC_CHARSET_GREEK
	| SC_CHARSET_HANGUL | SC_CHARSET_HEBREW | SC_CHARSET_JOHAB | SC_CHARSET_MAC
	| SC_CHARSET_OEM | SC_CHARSET_RUSSIAN | SC_CHARSET_SHIFTJIS | SC_CHARSET_SYMBOL
	| SC_CHARSET_THAI | SC_CHARSET_TURKISH | SC_CHARSET_VIETNAMESE

type command_msg = SCEN_KILLFOCUS | SCEN_SETFOCUS | SCEN_CHANGE

type edge_mode = EDGE_NONE | EDGE_LINE | EDGE_BACKGROUND

type eol_mode = SC_EOL_CR | SC_EOL_CRLF | SC_EOL_LF

type indicator_style = 
	| INDIC_PLAIN | INDIC_SQUIGGLE | INDIC_TT | INDIC_DIAGONAL
	| INDIC_STRIKE | INDIC_HIDDEN | INDIC_BOX | INDIC_ROUNDBOX

type key_code =
	| SCK_ADD | SCK_BACK | SCK_DELETE | SCK_DIVIDE | SCK_DOWN | SCK_END
	| SCK_ESCAPE | SCK_HOME | SCK_INSERT | SCK_LEFT | SCK_MENU | SCK_NEXT
	| SCK_PRIOR | SCK_RETURN | SCK_RIGHT | SCK_RWIN | SCK_SUBTRACT | SCK_TAB
	| SCK_UP | SCK_WIN

type key_modifier = (* SCMOD_NORM | *) SCMOD_SHIFT | SCMOD_CTRL | SCMOD_ALT

type key_definition = KeyDef of key_code * key_modifier list

type margin_type = SC_MARGIN_SYMBOL | SC_MARGIN_NUMBER | SC_MARGIN_BACK
	| SC_MARGIN_FORE

type marker_symbol =
	| SC_MARK_CIRCLE | SC_MARK_ROUNDRECT | SC_MARK_ARROW | SC_MARK_SMALLRECT
	| SC_MARK_SHORTARROW | SC_MARK_EMPTY | SC_MARK_ARROWDOWN | SC_MARK_MINUS
	| SC_MARK_PLUS | SC_MARK_VLINE | SC_MARK_LCORNER | SC_MARK_TCORNER
	| SC_MARK_BOXPLUS | SC_MARK_BOXPLUSCONNECTED | SC_MARK_BOXMINUS
	| SC_MARK_BOXMINUSCONNECTED | SC_MARK_LCORNERCURVE | SC_MARK_TCORNERCURVE
	| SC_MARK_CIRCLEPLUS | SC_MARK_CIRCLEPLUSCONNECTED | SC_MARK_CIRCLEMINUS
	| SC_MARK_CIRCLEMINUSCONNECTED | SC_MARK_BACKGROUND | SC_MARK_DOTDOTDOT
	| SC_MARK_ARROWS | SC_MARK_PIXMAP | SC_MARK_FULLRECT
	| SC_MARK_CHARACTER of char

type modification_type =
	| SC_MOD_INSERTTEXT | SC_MOD_DELETETEXT | SC_MOD_CHANGESTYLE
	| SC_MOD_CHANGEFOLD | SC_PERFORMED_USER | SC_PERFORMED_UNDO
	| SC_PERFORMED_REDO | SC_MULTISTEPUNDOREDO | SC_LASTSTEPINUNDOREDO
	| SC_MOD_CHANGEMARKER | SC_MOD_BEFOREINSERT | SC_MOD_BEFOREDELETE
	| SC_MULTILINEUNDOREDO | SC_STARTACTION (* | SC_MODEVENTMASKALL *)

(* N.B. - leave in NUMERIC order (and SCN_POSCHANGED is a "place-holder") *)
type notification_msg =
	| SCN_STYLENEEDED | SCN_CHARADDED | SCN_SAVEPOINTREACHED | SCN_SAVEPOINTLEFT
	| SCN_MODIFYATTEMPTRO | SCN_KEY | SCN_DOUBLECLICK | SCN_UPDATEUI
	| SCN_MODIFIED | SCN_MACRORECORD | SCN_MARGINCLICK | SCN_NEEDSHOWN
	| SCN_POSCHANGED | SCN_PAINTED | SCN_USERLISTSELECTION | SCN_URIDROPPED
	| SCN_DWELLSTART | SCN_DWELLEND | SCN_ZOOM | SCN_HOTSPOTCLICK
	| SCN_HOTSPOTDOUBLECLICK | SCN_CALLTIPCLICK | SCN_AUTOCSELECTION 

type search_flag =
	| SCFIND_MATCHCASE | SCFIND_WHOLEWORD | SCFIND_WORDSTART
	| SCFIND_REGEXP | SCFIND_POSIX

type selection_mode = SC_SEL_STREAM | SC_SEL_RECTANGLE | SC_SEL_LINES

type visible_policy = VISIBLE_SLOP | VISIBLE_STRICT

type visual_flag =
	| SC_WRAPVISUALFLAG_NONE | SC_WRAPVISUALFLAG_END | SC_WRAPVISUALFLAG_START

type visual_flag_loc =
	| SC_WRAPVISUALFLAGLOC_DEFAULT | SC_WRAPVISUALFLAGLOC_END_BY_TEXT
	| SC_WRAPVISUALFLAGLOC_START_BY_TEXT

type white_space_mode =
	| SCWS_INVISIBLE | SCWS_VISIBLEALWAYS | SCWS_VISIBLEAFTERINDENT

type wrap_mode = SC_WRAP_NONE | SC_WRAP_WORD | SC_WRAP_CHAR

(* "standard" Scintilla commands *)
type sci_cmd =
	| SCI_ADDREFDOCUMENT | SCI_ADDSTYLEDTEXT | SCI_ADDTEXT | SCI_ALLOCATE
	| SCI_APPENDTEXT | SCI_ASSIGNCMDKEY | SCI_AUTOCACTIVE | SCI_AUTOCCANCEL
	| SCI_AUTOCCOMPLETE | SCI_AUTOCGETAUTOHIDE | SCI_AUTOCGETCANCELATSTART
	| SCI_AUTOCGETCHOOSESINGLE | SCI_AUTOCGETCURRENT | SCI_AUTOCGETDROPRESTOFWORD
	| SCI_AUTOCGETIGNORECASE | SCI_AUTOCGETMAXHEIGHT | SCI_AUTOCGETMAXWIDTH
	| SCI_AUTOCGETSEPARATOR | SCI_AUTOCGETTYPESEPARATOR | SCI_AUTOCPOSSTART
	| SCI_AUTOCSELECT | SCI_AUTOCSETAUTOHIDE | SCI_AUTOCSETCANCELATSTART
	| SCI_AUTOCSETCHOOSESINGLE | SCI_AUTOCSETDROPRESTOFWORD | SCI_AUTOCSETFILLUPS
	| SCI_AUTOCSETIGNORECASE | SCI_AUTOCSETMAXHEIGHT | SCI_AUTOCSETMAXWIDTH
	| SCI_AUTOCSETSEPARATOR | SCI_AUTOCSETTYPESEPARATOR | SCI_AUTOCSHOW
	| SCI_AUTOCSTOPS | SCI_BEGINUNDOACTION | SCI_BRACEBADLIGHT
	| SCI_BRACEHIGHLIGHT | SCI_BRACEMATCH | SCI_CALLTIPACTIVE | SCI_CALLTIPCANCEL
	| SCI_CALLTIPPOSSTART | SCI_CALLTIPSETBACK | SCI_CALLTIPSETFORE
	| SCI_CALLTIPSETFOREHLT | SCI_CALLTIPSETHLT | SCI_CALLTIPSHOW
	| SCI_CALLTIPUSESTYLE | SCI_CANPASTE | SCI_CANREDO | SCI_CANUNDO
	| SCI_CHOOSECARETX | SCI_CLEAR | SCI_CLEARALL | SCI_CLEARALLCMDKEYS
	| SCI_CLEARCMDKEY | SCI_CLEARDOCUMENTSTYLE | SCI_CLEARREGISTEREDIMAGES
	| SCI_COLOURISE | SCI_CONVERTEOLS | SCI_COPY | SCI_COPYRANGE | SCI_COPYTEXT
	| SCI_CREATEDOCUMENT | SCI_CUT | SCI_DOCLINEFROMVISIBLE | SCI_EMPTYUNDOBUFFER
	| SCI_ENCODEDFROMUTF8 | SCI_ENDUNDOACTION | SCI_ENSUREVISIBLE
	| SCI_ENSUREVISIBLEENFORCEPOLICY | SCI_FINDCOLUMN | SCI_FINDTEXT
	| SCI_FORMATRANGE | SCI_GETANCHOR | SCI_GETBACKSPACEUNINDENTS
	| SCI_GETBUFFEREDDRAW | SCI_GETCARETFORE | SCI_GETCARETLINEBACK
	| SCI_GETCARETLINEBACKALPHA | SCI_GETCARETLINEVISIBLE | SCI_GETCARETPERIOD
	| SCI_GETCARETSTICKY | SCI_GETCARETWIDTH | SCI_GETCHARAT | SCI_GETCODEPAGE
	| SCI_GETCOLUMN | SCI_GETCONTROLCHARSYMBOL | SCI_GETCURLINE
	| SCI_GETCURRENTPOS | SCI_GETCURSOR | SCI_GETDIRECTFUNCTION
	| SCI_GETDIRECTPOINTER | SCI_GETDOCPOINTER | SCI_GETEDGECOLOUR
	| SCI_GETEDGECOLUMN | SCI_GETEDGEMODE | SCI_GETENDATLASTLINE
	| SCI_GETENDSTYLED | SCI_GETEOLMODE | SCI_GETFIRSTVISIBLELINE | SCI_GETFOCUS
	| SCI_GETFOLDEXPANDED | SCI_GETFOLDLEVEL | SCI_GETFOLDPARENT
	| SCI_GETHIGHLIGHTGUIDE | SCI_GETHSCROLLBAR | SCI_GETINDENT
	| SCI_GETINDENTATIONGUIDES | SCI_GETLASTCHILD | SCI_GETLAYOUTCACHE
	| SCI_GETLENGTH | SCI_GETLEXER | SCI_GETLINE | SCI_GETLINECOUNT
	| SCI_GETLINEENDPOSITION | SCI_GETLINEINDENTATION | SCI_GETLINEINDENTPOSITION
	| SCI_GETLINESELENDPOSITION | SCI_GETLINESELSTARTPOSITION | SCI_GETLINESTATE
	| SCI_GETLINEVISIBLE | SCI_GETMARGINLEFT | SCI_GETMARGINMASKN
	| SCI_GETMARGINRIGHT | SCI_GETMARGINSENSITIVEN | SCI_GETMARGINTYPEN
	| SCI_GETMARGINWIDTHN | SCI_GETMAXLINESTATE | SCI_GETMODEVENTMASK
	| SCI_GETMODIFY | SCI_GETMOUSEDOWNCAPTURES | SCI_GETMOUSEDWELLTIME
	| SCI_GETOVERTYPE | SCI_GETPASTECONVERTENDINGS | SCI_GETPRINTCOLOURMODE
	| SCI_GETPRINTMAGNIFICATION | SCI_GETPRINTWRAPMODE | SCI_GETPROPERTY
	| SCI_GETPROPERTYEXPANDED | SCI_GETPROPERTYINT | SCI_GETREADONLY
	| SCI_GETSCROLLWIDTH | SCI_GETSEARCHFLAGS | SCI_GETSELALPHA
	| SCI_GETSELECTIONEND | SCI_GETSELECTIONMODE | SCI_GETSELECTIONSTART
	| SCI_GETSELTEXT | SCI_GETSTATUS | SCI_GETSTYLEAT | SCI_GETSTYLEBITS
	| SCI_GETSTYLEBITSNEEDED | SCI_GETSTYLEDTEXT | SCI_GETTABINDENTS
	| SCI_GETTABWIDTH | SCI_GETTARGETEND | SCI_GETTARGETSTART | SCI_GETTEXT
	| SCI_GETTEXTLENGTH | SCI_GETTEXTRANGE | SCI_GETTWOPHASEDRAW
	| SCI_GETUNDOCOLLECTION | SCI_GETUSEPALETTE | SCI_GETUSETABS | SCI_GETVIEWEOL
	| SCI_GETVIEWWS | SCI_GETVSCROLLBAR | SCI_GETWRAPMODE | SCI_GETWRAPSTARTINDENT
	| SCI_GETWRAPVISUALFLAGS | SCI_GETWRAPVISUALFLAGSLOCATION | SCI_GETXOFFSET
	| SCI_GETZOOM | SCI_GOTOLINE | SCI_GOTOPOS | SCI_GRABFOCUS | SCI_HIDELINES
	| SCI_HIDESELECTION | SCI_INDICGETFORE | SCI_INDICGETSTYLE | SCI_INDICSETFORE
	| SCI_INDICSETSTYLE | SCI_INSERTTEXT | SCI_LINEFROMPOSITION | SCI_LINELENGTH
	| SCI_LINESCROLL | SCI_LINESJOIN | SCI_LINESONSCREEN | SCI_LINESSPLIT
	| SCI_LOADLEXERLIBRARY | SCI_MARKERADD | SCI_MARKERADDSET | SCI_MARKERDEFINE
	| SCI_MARKERDEFINEPIXMAP | SCI_MARKERDELETE | SCI_MARKERDELETEALL
	| SCI_MARKERDELETEHANDLE | SCI_MARKERGET | SCI_MARKERLINEFROMHANDLE
	| SCI_MARKERNEXT | SCI_MARKERPREVIOUS | SCI_MARKERSETALPHA | SCI_MARKERSETBACK
	| SCI_MARKERSETFORE | SCI_MOVECARETINSIDEVIEW | SCI_NULL | SCI_PASTE
	| SCI_POINTXFROMPOSITION | SCI_POINTYFROMPOSITION | SCI_POSITIONAFTER
	| SCI_POSITIONBEFORE | SCI_POSITIONFROMLINE | SCI_POSITIONFROMPOINT
	| SCI_POSITIONFROMPOINTCLOSE | SCI_REDO | SCI_REGISTERIMAGE
	| SCI_RELEASEDOCUMENT | SCI_REPLACESEL | SCI_REPLACETARGET
	| SCI_REPLACETARGETRE | SCI_SCROLLCARET | SCI_SEARCHANCHOR
	| SCI_SEARCHINTARGET | SCI_SEARCHNEXT | SCI_SEARCHPREV | SCI_SELECTALL
	| SCI_SELECTIONISRECTANGLE | SCI_SETANCHOR | SCI_SETBACKSPACEUNINDENTS
	| SCI_SETBUFFEREDDRAW | SCI_SETCARETFORE | SCI_SETCARETLINEBACK
	| SCI_SETCARETLINEBACKALPHA | SCI_SETCARETLINEVISIBLE | SCI_SETCARETPERIOD
	| SCI_SETCARETSTICKY | SCI_SETCARETWIDTH | SCI_SETCHARSDEFAULT
	| SCI_SETCODEPAGE | SCI_SETCONTROLCHARSYMBOL | SCI_SETCURRENTPOS
	| SCI_SETCURSOR | SCI_SETDOCPOINTER | SCI_SETEDGECOLOUR | SCI_SETEDGECOLUMN
	| SCI_SETEDGEMODE | SCI_SETENDATLASTLINE | SCI_SETEOLMODE | SCI_SETFOCUS
	| SCI_SETFOLDEXPANDED | SCI_SETFOLDFLAGS | SCI_SETFOLDLEVEL
	| SCI_SETFOLDMARGINCOLOUR | SCI_SETFOLDMARGINHICOLOUR | SCI_SETHIGHLIGHTGUIDE
	| SCI_SETHOTSPOTACTIVEBACK | SCI_SETHOTSPOTACTIVEFORE
	| SCI_SETHOTSPOTACTIVEUNDERLINE | SCI_SETHOTSPOTSINGLELINE | SCI_SETHSCROLLBAR
	| SCI_SETINDENT | SCI_SETINDENTATIONGUIDES | SCI_SETKEYWORDS
	| SCI_SETLAYOUTCACHE | SCI_SETLENGTHFORENCODE | SCI_SETLEXER
	| SCI_SETLEXERLANGUAGE | SCI_SETLINEINDENTATION | SCI_SETLINESTATE
	| SCI_SETMARGINLEFT | SCI_SETMARGINMASKN | SCI_SETMARGINRIGHT
	| SCI_SETMARGINSENSITIVEN | SCI_SETMARGINTYPEN | SCI_SETMARGINWIDTHN
	| SCI_SETMODEVENTMASK | SCI_SETMOUSEDOWNCAPTURES | SCI_SETMOUSEDWELLTIME
	| SCI_SETOVERTYPE | SCI_SETPASTECONVERTENDINGS | SCI_SETPRINTCOLOURMODE
	| SCI_SETPRINTMAGNIFICATION | SCI_SETPRINTWRAPMODE | SCI_SETPROPERTY
	| SCI_SETREADONLY | SCI_SETSAVEPOINT | SCI_SETSCROLLWIDTH | SCI_SETSEARCHFLAGS
	| SCI_SETSEL | SCI_SETSELALPHA | SCI_SETSELBACK | SCI_SETSELECTIONEND
	| SCI_SETSELECTIONMODE | SCI_SETSELECTIONSTART | SCI_SETSELFORE
	| SCI_SETSTATUS | SCI_SETSTYLEBITS | SCI_SETSTYLING | SCI_SETSTYLINGEX
	| SCI_SETTABINDENTS | SCI_SETTABWIDTH | SCI_SETTARGETEND | SCI_SETTARGETSTART
	| SCI_SETTEXT | SCI_SETTWOPHASEDRAW | SCI_SETUNDOCOLLECTION
	| SCI_SETUSEPALETTE | SCI_SETUSETABS | SCI_SETVIEWEOL | SCI_SETVIEWWS
	| SCI_SETVISIBLEPOLICY | SCI_SETVSCROLLBAR | SCI_SETWHITESPACEBACK
	| SCI_SETWHITESPACECHARS | SCI_SETWHITESPACEFORE | SCI_SETWORDCHARS
	| SCI_SETWRAPMODE | SCI_SETWRAPSTARTINDENT | SCI_SETWRAPVISUALFLAGS
	| SCI_SETWRAPVISUALFLAGSLOCATION | SCI_SETXCARETPOLICY | SCI_SETXOFFSET
	| SCI_SETYCARETPOLICY | SCI_SETZOOM | SCI_SHOWLINES | SCI_STARTRECORD
	| SCI_STARTSTYLING | SCI_STOPRECORD | SCI_STYLECLEARALL
	| SCI_STYLERESETDEFAULT | SCI_STYLESETBACK | SCI_STYLESETBOLD
	| SCI_STYLESETCASE | SCI_STYLESETCHANGEABLE | SCI_STYLESETCHARACTERSET
	| SCI_STYLESETEOLFILLED | SCI_STYLESETFONT | SCI_STYLESETFORE
	| SCI_STYLESETHOTSPOT | SCI_STYLESETITALIC | SCI_STYLESETSIZE
	| SCI_STYLESETUNDERLINE | SCI_STYLESETVISIBLE | SCI_TARGETASUTF8
	| SCI_TARGETFROMSELECTION | SCI_TEXTHEIGHT | SCI_TEXTWIDTH
	| SCI_TOGGLECARETSTICKY | SCI_TOGGLEFOLD | SCI_UNDO | SCI_USEPOPUP
	| SCI_USERLISTSHOW | SCI_VISIBLEFROMDOCLINE | SCI_WORDENDPOSITION
	| SCI_WORDSTARTPOSITION | SCI_WRAPCOUNT | SCI_ZOOMIN | SCI_ZOOMOUT

(* "keyboard" Scintilla commands *)
type key_cmd =
	| SCI_BACKTAB | SCI_CANCEL | SCI_CHARLEFT | SCI_CHARLEFTEXTEND
	| SCI_CHARLEFTRECTEXTEND | SCI_CHARRIGHT | SCI_CHARRIGHTEXTEND
	| SCI_CHARRIGHTRECTEXTEND | SCI_DELETEBACK | SCI_DELETEBACKNOTLINE
	| SCI_DELLINELEFT | SCI_DELLINERIGHT | SCI_DELWORDLEFT | SCI_DELWORDRIGHT
	| SCI_DOCUMENTEND | SCI_DOCUMENTENDEXTEND | SCI_DOCUMENTSTART
	| SCI_DOCUMENTSTARTEXTEND | SCI_EDITTOGGLEOVERTYPE | SCI_FORMFEED | SCI_HOME
	| SCI_HOMEDISPLAY | SCI_HOMEDISPLAYEXTEND | SCI_HOMEEXTEND
	| SCI_HOMERECTEXTEND | SCI_HOMEWRAP | SCI_HOMEWRAPEXTEND | SCI_LINECOPY
	| SCI_LINECUT | SCI_LINEDELETE | SCI_LINEDOWN | SCI_LINEDOWNEXTEND
	| SCI_LINEDOWNRECTEXTEND | SCI_LINEDUPLICATE | SCI_LINEEND
	| SCI_LINEENDDISPLAY | SCI_LINEENDDISPLAYEXTEND | SCI_LINEENDEXTEND
	| SCI_LINEENDRECTEXTEND | SCI_LINEENDWRAP | SCI_LINEENDWRAPEXTEND
	| SCI_LINESCROLLDOWN | SCI_LINESCROLLUP | SCI_LINETRANSPOSE | SCI_LINEUP
	| SCI_LINEUPEXTEND | SCI_LINEUPRECTEXTEND | SCI_LOWERCASE | SCI_NEWLINE
	| SCI_PAGEDOWN | SCI_PAGEDOWNEXTEND | SCI_PAGEDOWNRECTEXTEND | SCI_PAGEUP
	| SCI_PAGEUPEXTEND | SCI_PAGEUPRECTEXTEND | SCI_PARADOWN | SCI_PARADOWNEXTEND
	| SCI_PARAUP | SCI_PARAUPEXTEND | SCI_SELECTIONDUPLICATE
	| SCI_STUTTEREDPAGEDOWN | SCI_STUTTEREDPAGEDOWNEXTEND | SCI_STUTTEREDPAGEUP
	| SCI_STUTTEREDPAGEUPEXTEND | SCI_TAB | SCI_UPPERCASE | SCI_VCHOME
	| SCI_VCHOMEEXTEND | SCI_VCHOMERECTEXTEND | SCI_VCHOMEWRAP
	| SCI_VCHOMEWRAPEXTEND | SCI_WORDLEFT | SCI_WORDLEFTEND
	| SCI_WORDLEFTENDEXTEND | SCI_WORDLEFTEXTEND | SCI_WORDPARTLEFT
	| SCI_WORDPARTLEFTEXTEND | SCI_WORDPARTRIGHT | SCI_WORDPARTRIGHTEXTEND
	| SCI_WORDRIGHT | SCI_WORDRIGHTEND | SCI_WORDRIGHTENDEXTEND
	| SCI_WORDRIGHTEXTEND

(* N.B. - these are ONLY created by ml_scn_from_boxed_pointer *)
type scnotification = private {
	hwndFrom: opaque_ptr; idFrom: int; code: notification_msg;
	position: int;
	ch: int;
	modifiers: key_modifier list;
	modificationType: modification_type list;
	text: string; length: int;
	linesAdded: int;
	message: int;
	wParam: opaque_ptr; lParam: opaque_ptr;
	line: int;
	foldLevelNow: int; foldLevelPrev: int;
	margin: int;
	x: int; y: int
}

(* mapping of NON-typed/integer Scintilla constants *)
let sci_SC_CP_DBCS = 1
let sci_SC_CP_UTF8 = 65001

let sci_SC_CURSORNORMAL = -1
let sci_SC_CURSORWAIT = 4

let sci_SC_TIME_FOREVER = 10000000

let sci_INVALID_POSITION = -1

let sci_SC_MARKNUM_FOLDEREND = 25
let sci_SC_MARKNUM_FOLDEROPENMID = 26
let sci_SC_MARKNUM_FOLDERMIDTAIL = 27
let sci_SC_MARKNUM_FOLDERTAIL = 28
let sci_SC_MARKNUM_FOLDERSUB = 29
let sci_SC_MARKNUM_FOLDER = 30
let sci_SC_MARKNUM_FOLDEROPEN = 31

let sci_SC_MASK_FOLDERS = 0xFE000000l

let sci_SC_FOLDLEVELBASE = 0x400
let sci_SC_FOLDLEVELWHITEFLAG = 0x1000
let sci_SC_FOLDLEVELHEADERFLAG = 0x2000
let sci_SC_FOLDLEVELBOXHEADERFLAG = 0x4000
let sci_SC_FOLDLEVELBOXFOOTERFLAG = 0x8000
let sci_SC_FOLDLEVELCONTRACTED = 0x10000
let sci_SC_FOLDLEVELUNINDENT = 0x20000
let sci_SC_FOLDLEVELNUMBERMASK = 0x0FFF

let sci_STYLE_DEFAULT = 32
let sci_STYLE_LINENUMBER = 33
let sci_STYLE_BRACELIGHT = 34
let sci_STYLE_BRACEBAD = 35
let sci_STYLE_CONTROLCHAR = 36
let sci_STYLE_INDENTGUIDE = 37
let sci_STYLE_CALLTIP = 38
let sci_STYLE_LASTPREDEFINED = 39
let sci_STYLE_MAX = 127

let sci_INDIC0_MASK = 0x20
let sci_INDIC1_MASK = 0x40
let sci_INDIC2_MASK = 0x80
let sci_INDICS_MASK = 0xE0

let sci_SC_ALPHA_TRANSPARENT = 0
let sci_SC_ALPHA_OPAQUE = 255
let sci_SC_ALPHA_NOALPHA = 256

(* "type conversion" *)
external int_of_caret_policy_list: caret_policy list -> int =
	"ml_int_of_caret_policy_list" "noalloc"
external int_of_char_set: char_set -> int = "ml_int_of_char_set" "noalloc"
external int_of_key_cmd: key_cmd -> int = "ml_int_of_key_cmd" "noalloc"
external int_of_key_code: key_code -> int = "ml_int_of_key_code" "noalloc"
external int_of_key_modifier_list: key_modifier list -> int =
	"ml_int_of_key_modifier_list" "noalloc"
external int_of_marker_symbol: marker_symbol -> int =
	"ml_int_of_marker_symbol" "noalloc"
external int_of_modification_type_list: modification_type list -> int =
	"ml_int_of_modification_type_list" "noalloc"
external int_of_search_flag_list: search_flag list -> int =
	"ml_int_of_search_flag_list" "noalloc"
external int_of_visible_policy_list: visible_policy list -> int =
	"ml_int_of_visible_policy_list" "noalloc"
external int_of_visual_flag_list: visual_flag list -> int =
	"ml_int_of_visual_flag_list" "noalloc"
external int_of_visual_flag_loc_list: visual_flag_loc list -> int =
	"ml_int_of_visual_flag_loc_list" "noalloc"
external modification_type_of_int: int -> modification_type = "%identity"
external search_flag_of_int: int -> search_flag = "ml_search_flag_of_int"
external visual_flag_of_int: int -> visual_flag = "%identity"
external visual_flag_loc_of_int: int -> visual_flag_loc = "%identity"

external cm_from_int: int -> command_msg = "ml_cm_from_int"
external scn_from_boxed_pointer: Gpointer.boxed option -> scnotification =
	"ml_scn_from_boxed_pointer"

(* low-level widget interface - GTK support *)
external sci_new: unit -> scintilla obj = "ml_sci_new"
external sci_get_type: unit -> Gobject.g_type = "ml_sci_get_type"
external sci_release_resources: unit -> unit = "ml_sci_release_resources"
external sci_set_id: scintilla obj -> int -> unit = "ml_sci_set_id" "noalloc"

(* low-level widget interface - internal Scintilla "messaging" *)
external sci_send_message_Ciu: scintilla obj -> sci_cmd -> cache_mode -> int -> unit =
	"ml_sci_send_message_iiu" "noalloc"
external sci_send_message_Eiu: scintilla obj -> sci_cmd -> edge_mode -> int -> unit =
	"ml_sci_send_message_iiu" "noalloc"
external sci_send_message_Liu: scintilla obj -> sci_cmd -> eol_mode -> int -> unit =
	"ml_sci_send_message_iiu" "noalloc"
external sci_send_message_Siu: scintilla obj -> sci_cmd -> selection_mode -> int -> unit =
	"ml_sci_send_message_iiu" "noalloc"
external sci_send_message_Uiu: scintilla obj -> sci_cmd -> wrap_mode -> int -> unit =
	"ml_sci_send_message_iiu" "noalloc"
external sci_send_message_Wiu: scintilla obj -> sci_cmd -> white_space_mode -> int -> unit =
	"ml_sci_send_message_iiu" "noalloc"
external sci_send_message_biu: scintilla obj -> sci_cmd -> bool -> int -> unit =
	"ml_sci_send_message_iiu" "noalloc"
external sci_send_message_i3i: scintilla obj -> sci_cmd -> int -> int32 -> int =
	"ml_sci_send_message_i3i" "noalloc"
external sci_send_message_i3u: scintilla obj -> sci_cmd -> int -> int32 -> unit =
	"ml_sci_send_message_i3u" "noalloc"
external sci_send_message_iAu: scintilla obj -> sci_cmd -> int -> case_mode -> unit =
	"ml_sci_send_message_iiu" "noalloc"
external sci_send_message_iIu: scintilla obj -> sci_cmd -> int -> indicator_style -> unit =
	"ml_sci_send_message_iiu" "noalloc"
external sci_send_message_iMu: scintilla obj -> sci_cmd -> int -> margin_type -> unit =
	"ml_sci_send_message_iiu" "noalloc"
external sci_send_message_ibi: scintilla obj -> sci_cmd -> int -> bool -> int =
	"ml_sci_send_message_iii" "noalloc"
external sci_send_message_ibu: scintilla obj -> sci_cmd -> int -> bool -> unit =
	"ml_sci_send_message_iiu" "noalloc"
external sci_send_message_idu: scintilla obj -> sci_cmd -> int -> document -> unit =
	"ml_sci_send_message_inu" "noalloc"
external sci_send_message_ii3: scintilla obj -> sci_cmd -> int -> int -> int32 =
	"ml_sci_send_message_ii3" "noalloc"
external sci_send_message_iiC: scintilla obj -> sci_cmd -> int -> int -> cache_mode =
	"ml_sci_send_message_iii" "noalloc"
external sci_send_message_iiE: scintilla obj -> sci_cmd -> int -> int -> edge_mode =
	"ml_sci_send_message_iii" "noalloc"
external sci_send_message_iiI: scintilla obj -> sci_cmd -> int -> int -> indicator_style =
	"ml_sci_send_message_iii" "noalloc"
external sci_send_message_iiL: scintilla obj -> sci_cmd -> int -> int -> eol_mode =
	"ml_sci_send_message_iii" "noalloc"
external sci_send_message_iiM: scintilla obj -> sci_cmd -> int -> int -> margin_type =
	"ml_sci_send_message_iii" "noalloc"
external sci_send_message_iiS: scintilla obj -> sci_cmd -> int -> int -> selection_mode =
	"ml_sci_send_message_iii" "noalloc"
external sci_send_message_iiU: scintilla obj -> sci_cmd -> int -> int -> wrap_mode =
	"ml_sci_send_message_iii" "noalloc"
external sci_send_message_iiW: scintilla obj -> sci_cmd -> int -> int -> white_space_mode =
	"ml_sci_send_message_iii" "noalloc"
external sci_send_message_iib: scintilla obj -> sci_cmd -> int -> int -> bool =
	"ml_sci_send_message_iii" "noalloc"
external sci_send_message_iic: scintilla obj -> sci_cmd -> int -> int -> char =
	"ml_sci_send_message_iii" "noalloc"
external sci_send_message_iid: scintilla obj -> sci_cmd -> int -> int -> document =
	"ml_sci_send_message_iin" "noalloc"
external sci_send_message_iii: scintilla obj -> sci_cmd -> int -> int -> int =
	"ml_sci_send_message_iii" "noalloc"
external sci_send_message_iiu: scintilla obj -> sci_cmd -> int -> int -> unit =
	"ml_sci_send_message_iiu" "noalloc"
external sci_send_message_irsr: scintilla obj -> sci_cmd -> int -> text_range -> string -> text_range =
	"ml_sci_send_message_irsr"
external sci_send_message_isi: scintilla obj -> sci_cmd -> int -> string -> int =
	"ml_sci_send_message_isi" "noalloc"
external sci_send_message_isu: scintilla obj -> sci_cmd -> int -> string -> unit =
	"ml_sci_send_message_isu" "noalloc"
external sci_send_message_rsi: scintilla obj -> sci_cmd -> text_range -> string -> int =
	"ml_sci_send_message_rsi" "noalloc"
external sci_send_message_sii: scintilla obj -> sci_cmd -> string -> int -> int =
	"ml_sci_send_message_sii" "noalloc"
external sci_send_message_ssi: scintilla obj -> sci_cmd -> string -> string -> int =
	"ml_sci_send_message_ssi" "noalloc"
external sci_send_message_ssu: scintilla obj -> sci_cmd -> string -> string -> unit =
	"ml_sci_send_message_ssu" "noalloc"

external sci_send_message_key: scintilla obj -> key_cmd -> unit =
	"ml_sci_send_message_key" "noalloc"

let int_of_key_definition = function
	| KeyDef (k, m) -> (int_of_key_modifier_list m) lsl 16 + (int_of_key_code k)

(* EXPLODE positive int flags collection into flag list *)
let explode_all f = function
	| 0 -> []
	| m ->
		let rec test m' =
			(* stop on "match not possible" or "overflow" *)
			if m' > m || m' <= 0 then []
			else
				if m' land m == 0 then
					test (m' lsl 1)
				else
					f m' :: test (m' lsl 1) in
		test 1

let explode_all_exp f = function
	| 0 -> []
	| m ->
		let rec test e =
			let m' = 1 lsl e in
			(* stop on "match not possible" or "overflow" *)
			if m' > m || m' <= 0 then []
			else
				if m' land m == 0 then
					test (succ e)
				else
					f e :: test (succ e) in
		test 0

(* define the "signals" interface for the Scintilla widget *)
(* N.B. - nearly everything of interest will appear as "sci-notify" signals *)
module S = struct
	open GtkSignal
	open Gobject.Data
	(* callback: int -> command_msg -> Gpointer.boxed option -> unit *)
	let marshal_command f argv =
		let g f cm_id bp =
			f (cm_id land 0xffff) (cm_from_int cm_id) bp in
		marshal2 int pointer "command" (g f) argv
	let command = {name="command"; classe=`scintilla; marshaller=marshal_command}
	(* callback: int -> scnotification -> unit *)
	let marshal_notify f argv =
		let g f id bp =
			f id (scn_from_boxed_pointer bp) in
		marshal2 int pointer "sci-notify" (g f) argv
	let notify = {name="sci-notify"; classe=`scintilla; marshaller=marshal_notify}
end

class virtual scintilla_sigs = object (self)
	method private virtual connect: 'b. ('a,'b) GtkSignal.t -> callback:'b -> GtkSignal.id
	method command = self#connect S.command
	method notify = self#connect S.notify
end

class scintilla_signals obj = object (self)
	inherit widget_signals_impl (obj: [> scintilla] obj)
	inherit scintilla_sigs
end

(* the Scintilla editor widget wrapper class *)
class editor obj = object (self)
(*	inherit widget obj*)
	inherit [scintilla] widget_impl obj
	method connect = new scintilla_signals obj

	method set_id id = sci_set_id obj id

	(* Text retrieval and modification *)
	method get_text len =
		if len <= 0 then ""
		else
			let buf = String.create (pred len) in
			sci_send_message_isi obj SCI_GETTEXT len buf; buf
	method set_text buf = sci_send_message_isu obj SCI_SETTEXT 0 buf
	method set_save_point () = sci_send_message_iiu obj SCI_SETSAVEPOINT 0 0
	method get_line line =
		let n = self#line_length line in
		if n <= 0 then ""
		else
			let buf = String.create n in
			sci_send_message_isi obj SCI_GETLINE line buf; buf
	method replace_sel buf = sci_send_message_isu obj SCI_REPLACESEL 0 buf
	method set_readonly mode = sci_send_message_biu obj SCI_SETREADONLY mode 0
	method get_readonly = sci_send_message_iib obj SCI_GETREADONLY 0 0
	method get_text_range tr =
		let n = (if tr.cpMax == -1 then self#get_text_length else tr.cpMax) - tr.cpMin in
		let buf = String.create n in
		sci_send_message_rsi obj SCI_GETTEXTRANGE tr buf; buf
	method allocate len = sci_send_message_iiu obj SCI_ALLOCATE len 0
	method add_text len buf = sci_send_message_isu obj SCI_ADDTEXT len buf
	method add_styled_text len buf = sci_send_message_isu obj SCI_ADDSTYLEDTEXT len buf
	method append_text len buf = sci_send_message_isu obj SCI_APPENDTEXT len buf
	method insert_text pos buf = sci_send_message_isu obj SCI_INSERTTEXT pos buf
	method clear_all () = sci_send_message_iiu obj SCI_CLEARALL 0 0
	method clear_document_style () = sci_send_message_iiu obj SCI_CLEARDOCUMENTSTYLE 0 0
	method get_char_at pos = sci_send_message_iic obj SCI_GETCHARAT pos 0
	method get_style_at pos = (sci_send_message_iii obj SCI_GETSTYLEAT pos 0) land 0xff
	method get_styled_text tr =
		let n = tr.cpMax - tr.cpMin in
		let buf = String.create (n * 2) in
		sci_send_message_rsi obj SCI_GETSTYLEDTEXT tr buf; buf
	method set_style_bits n = sci_send_message_iiu obj SCI_SETSTYLEBITS n 0
	method get_style_bits = sci_send_message_iii obj SCI_GETSTYLEBITS 0 0
	(* the next 3 methods are GTK-specific *)
	method target_as_utf8 =
		let n = (self#get_target_end - self#get_target_start) * 3 in
		let buf = String.create n in
		let n' = sci_send_message_isi obj SCI_TARGETASUTF8 0 buf in
		if n' < n then String.sub buf 0 n' else buf
	method encoded_from_utf8 buf =
		let n = String.length buf in
		let buf' = String.create n in
		let n' = sci_send_message_ssi obj SCI_ENCODEDFROMUTF8 buf buf' in
		if n' < n then String.sub buf' 0 n' else buf'
	method set_length_for_encode n = sci_send_message_iiu obj SCI_SETLENGTHFORENCODE n 0

	(* Searching and Replacing *)
	method find_text ?(flags=[]) tr buf =
		sci_send_message_irsr obj SCI_FINDTEXT (int_of_search_flag_list flags) tr buf
	method search_anchor () = sci_send_message_iiu obj SCI_SEARCHANCHOR 0 0
	method search_next ?(flags=[]) buf =
		sci_send_message_isi obj SCI_SEARCHNEXT (int_of_search_flag_list flags) buf
	method search_prev ?(flags=[]) buf =
		sci_send_message_isi obj SCI_SEARCHPREV (int_of_search_flag_list flags) buf
	method set_target_start pos = sci_send_message_iiu obj SCI_SETTARGETSTART pos 0
	method get_target_start = sci_send_message_iii obj SCI_GETTARGETSTART 0 0
	method set_target_end pos = sci_send_message_iiu obj SCI_SETTARGETEND pos 0
	method get_target_end = sci_send_message_iii obj SCI_GETTARGETEND 0 0
	method target_from_selection () =
		sci_send_message_iiu obj SCI_TARGETFROMSELECTION 0 0
	method set_search_flags ?(flags=[]) () =
		sci_send_message_iiu obj SCI_SETSEARCHFLAGS (int_of_search_flag_list flags) 0
	method get_search_flags =
		explode_all search_flag_of_int (sci_send_message_iii obj SCI_GETSEARCHFLAGS 0 0)
	method search_in_target len buf =
		sci_send_message_isi obj SCI_SEARCHINTARGET len buf
	method replace_target len buf = sci_send_message_isi obj SCI_REPLACETARGET len buf
	method replace_target_re len buf =
		sci_send_message_isi obj SCI_REPLACETARGETRE len buf

	(* Overtype *)
	method set_overtype mode = sci_send_message_biu obj SCI_SETOVERTYPE mode 0
	method get_overtype = sci_send_message_iib obj SCI_GETOVERTYPE 0 0

	(* Cut, copy and paste *)
	method cut () = sci_send_message_iiu obj SCI_CUT 0 0
	method copy () = sci_send_message_iiu obj SCI_COPY 0 0
	method paste () = sci_send_message_iiu obj SCI_PASTE 0 0
	method clear () = sci_send_message_iiu obj SCI_CLEAR 0 0
	method can_paste = sci_send_message_iii obj SCI_CANPASTE 0 0
	method copy_range pos1 pos2 = sci_send_message_iiu obj SCI_COPYRANGE pos1 pos2
	method copy_text len buf = sci_send_message_isu obj SCI_COPYTEXT len buf
	method set_paste_convert_endings mode =
		sci_send_message_biu obj SCI_SETPASTECONVERTENDINGS mode 0
	method get_paste_convert_endings =
		sci_send_message_iib obj SCI_GETPASTECONVERTENDINGS 0 0

	(* Error handling *)
	method set_status mode = sci_send_message_iiu obj SCI_SETSTATUS mode 0
	method get_status = sci_send_message_iii obj SCI_GETSTATUS 0 0

	(* Undo and Redo *)
	method undo () = sci_send_message_iiu obj SCI_UNDO 0 0
	method can_undo = sci_send_message_iii obj SCI_CANUNDO 0 0
	method redo () = sci_send_message_iiu obj SCI_REDO 0 0
	method can_redo = sci_send_message_iii obj SCI_CANREDO 0 0
	method empty_undo_buffer () = sci_send_message_iiu obj SCI_EMPTYUNDOBUFFER 0 0
	method set_undo_collection mode =
		sci_send_message_biu obj SCI_SETUNDOCOLLECTION mode 0
	method get_undo_collection = sci_send_message_iib obj SCI_GETUNDOCOLLECTION 0 0
	method begin_undo_action () = sci_send_message_iiu obj SCI_BEGINUNDOACTION 0 0
	method end_undo_action () = sci_send_message_iiu obj SCI_ENDUNDOACTION 0 0

	(* Selection and information *)
	method get_text_length = sci_send_message_iii obj SCI_GETTEXTLENGTH 0 0
	method get_length = sci_send_message_iii obj SCI_GETLENGTH 0 0
	method get_line_count = sci_send_message_iii obj SCI_GETLINECOUNT 0 0
	method get_first_visible_line = sci_send_message_iii obj SCI_GETFIRSTVISIBLELINE 0 0
	method lines_on_screen = sci_send_message_iii obj SCI_LINESONSCREEN 0 0
	method get_modify = sci_send_message_iii obj SCI_GETMODIFY 0 0
	method set_sel pos1 pos2 = sci_send_message_iiu obj SCI_SETSEL pos1 pos2
	method goto_pos pos = sci_send_message_iiu obj SCI_GOTOPOS pos 0
	method goto_line line = sci_send_message_iiu obj SCI_GOTOLINE line 0
	method set_current_pos pos = sci_send_message_iiu obj SCI_SETCURRENTPOS pos 0
	method get_current_pos = sci_send_message_iii obj SCI_GETCURRENTPOS 0 0
	method set_anchor pos = sci_send_message_iiu obj SCI_SETANCHOR pos 0
	method get_anchor = sci_send_message_iii obj SCI_GETANCHOR 0 0
	method set_selection_start pos =
		sci_send_message_iiu obj SCI_SETSELECTIONSTART pos 0
	method get_selection_start = sci_send_message_iii obj SCI_GETSELECTIONSTART 0 0
	method set_selection_end pos = sci_send_message_iiu obj SCI_SETSELECTIONEND pos 0
	method get_selection_end = sci_send_message_iii obj SCI_GETSELECTIONEND 0 0
	method select_all () = sci_send_message_iiu obj SCI_SELECTALL 0 0
	method line_from_position pos = sci_send_message_iii obj SCI_LINEFROMPOSITION pos 0
	method position_from_line line =
		sci_send_message_iii obj SCI_POSITIONFROMLINE line 0
	method get_line_end_position line =
		sci_send_message_iii obj SCI_GETLINEENDPOSITION line 0
	method line_length line = sci_send_message_iii obj SCI_LINELENGTH line 0
	method get_sel_text =
		let n = self#get_selection_end - self#get_selection_start in
		if n <= 0 then ""
		else
			let buf = String.create n in
			sci_send_message_isi obj SCI_GETSELTEXT 0 buf; buf
	method get_cur_line =
		let n = self#line_length (self#line_from_position self#get_current_pos) in
		if n <= 0 then ""
		else
			let buf = String.create n in
			sci_send_message_isi obj SCI_GETCURLINE (succ n) buf; buf
	method selection_is_rectangle =
		sci_send_message_iii obj SCI_SELECTIONISRECTANGLE 0 0
	method set_selection_mode mode =
		sci_send_message_Siu obj SCI_SETSELECTIONMODE mode 0
	method get_selection_mode = sci_send_message_iiS obj SCI_GETSELECTIONMODE 0 0
	method get_line_sel_start_position line =
		sci_send_message_iii obj SCI_GETLINESELSTARTPOSITION line 0
	method get_line_sel_end_position line =
		sci_send_message_iii obj SCI_GETLINESELENDPOSITION line 0
	method move_caret_inside_view () =
		sci_send_message_iiu obj SCI_MOVECARETINSIDEVIEW 0 0
	method word_end_position pos mode =
		sci_send_message_ibi obj SCI_WORDENDPOSITION pos mode
	method word_start_position pos mode =
		sci_send_message_ibi obj SCI_WORDSTARTPOSITION pos mode
	method position_before pos = sci_send_message_iii obj SCI_POSITIONBEFORE pos 0
	method position_after pos = sci_send_message_iii obj SCI_POSITIONAFTER pos 0
	method text_width pos buf = sci_send_message_isi obj SCI_TEXTWIDTH pos buf
	method text_height pos = sci_send_message_iii obj SCI_TEXTHEIGHT pos 0
	method get_column pos = sci_send_message_iii obj SCI_GETCOLUMN pos 0
	method find_column pos1 pos2 = sci_send_message_iii obj SCI_FINDCOLUMN pos1 pos2
	method position_from_point x y = sci_send_message_iii obj SCI_POSITIONFROMPOINT x y
	method position_from_point_close x y =
		sci_send_message_iii obj SCI_POSITIONFROMPOINTCLOSE x y
	method point_x_from_position pos =
		sci_send_message_iii obj SCI_POINTXFROMPOSITION 0 pos
	method point_y_from_position pos =
		sci_send_message_iii obj SCI_POINTYFROMPOSITION 0 pos
	method hide_selection mode = sci_send_message_biu obj SCI_HIDESELECTION mode 0
	method choose_caret_x () = sci_send_message_iiu obj SCI_CHOOSECARETX 0 0

	(* Scrolling and automatic scrolling *)
	method line_scroll col line = sci_send_message_iiu obj SCI_LINESCROLL col line
	method scroll_caret () = sci_send_message_iiu obj SCI_SCROLLCARET 0 0
	method set_x_caret_policy ?(flags=[]) x =
		sci_send_message_iiu obj SCI_SETXCARETPOLICY (int_of_caret_policy_list flags) x
	method set_y_caret_policy ?(flags=[]) y =
		sci_send_message_iiu obj SCI_SETYCARETPOLICY (int_of_caret_policy_list flags) y
	method set_visible_policy ?(flags=[]) y =
		sci_send_message_iiu obj SCI_SETVISIBLEPOLICY (int_of_visible_policy_list flags) y
	method set_h_scrollbar mode = sci_send_message_biu obj SCI_SETHSCROLLBAR mode 0
	method get_h_scrollbar = sci_send_message_iib obj SCI_GETHSCROLLBAR 0 0
	method set_v_scrollbar mode = sci_send_message_biu obj SCI_SETVSCROLLBAR mode 0
	method get_v_scrollbar = sci_send_message_iib obj SCI_GETVSCROLLBAR 0 0
	method set_x_offset x = sci_send_message_iiu obj SCI_SETXOFFSET x 0
	method get_x_offset = sci_send_message_iii obj SCI_GETXOFFSET 0 0
	method set_scroll_width x = sci_send_message_iiu obj SCI_SETSCROLLWIDTH x 0
	method get_scroll_width = sci_send_message_iii obj SCI_GETSCROLLWIDTH 0 0
	method set_end_at_last_line mode =
		sci_send_message_biu obj SCI_SETENDATLASTLINE mode 0
	method get_end_at_last_line = sci_send_message_iib obj SCI_GETENDATLASTLINE 0 0

	(* White space *)
	method set_view_ws mode = sci_send_message_Wiu obj SCI_SETVIEWWS mode 0
	method get_view_ws = sci_send_message_iiW obj SCI_GETVIEWWS 0 0
	method set_white_space_fore mode rgb =
		sci_send_message_biu obj SCI_SETWHITESPACEFORE mode rgb
	method set_white_space_back mode rgb =
		sci_send_message_biu obj SCI_SETWHITESPACEBACK mode rgb

	(* Cursor *)
	method set_cursor mode = sci_send_message_iiu obj SCI_SETCURSOR mode 0
	method get_cursor = sci_send_message_iii obj SCI_GETCURSOR 0 0

	(* Mouse capture *)
	method set_mouse_down_captures mode =
		sci_send_message_biu obj SCI_SETMOUSEDOWNCAPTURES mode 0
	method get_mouse_down_captures =
		sci_send_message_iib obj SCI_GETMOUSEDOWNCAPTURES 0 0

	(* Line endings *)
	method set_eol_mode mode = sci_send_message_Liu obj SCI_SETEOLMODE mode 0
	method get_eol_mode = sci_send_message_iiL obj SCI_GETEOLMODE 0 0
	method convert_eols mode = sci_send_message_Liu obj SCI_CONVERTEOLS mode 0
	method set_view_eol mode = sci_send_message_biu obj SCI_SETVIEWEOL mode 0
	method get_view_eol = sci_send_message_iib obj SCI_GETVIEWEOL 0 0

	(* Styling *)
	method get_end_styled = sci_send_message_iii obj SCI_GETENDSTYLED 0 0
	method start_styling pos mode =
		sci_send_message_iiu obj SCI_STARTSTYLING pos mode
	method set_styling len mode = sci_send_message_iiu obj SCI_SETSTYLING len mode
	method set_styling_ex len buf = sci_send_message_isu obj SCI_SETSTYLINGEX len buf
	method set_line_state line mode =
		sci_send_message_iiu obj SCI_SETLINESTATE line mode
	method get_line_state line = sci_send_message_iii obj SCI_GETLINESTATE line 0
	method get_max_line_state = sci_send_message_iii obj SCI_GETMAXLINESTATE 0 0

	(* Style definition *)
	method style_reset_default () = sci_send_message_iiu obj SCI_STYLERESETDEFAULT 0 0
	method style_clear_all () = sci_send_message_iiu obj SCI_STYLECLEARALL 0 0
	method style_set_font sel buf = sci_send_message_isu obj SCI_STYLESETFONT sel buf
	method style_set_size sel n = sci_send_message_iiu obj SCI_STYLESETSIZE sel n
	method style_set_bold sel mode = sci_send_message_ibu obj SCI_STYLESETBOLD sel mode
	method style_set_italic sel mode =
		sci_send_message_ibu obj SCI_STYLESETITALIC sel mode
	method style_set_underline sel mode =
		sci_send_message_ibu obj SCI_STYLESETUNDERLINE sel mode
	method style_set_fore sel rgb = sci_send_message_iiu obj SCI_STYLESETFORE sel rgb
	method style_set_back sel rgb = sci_send_message_iiu obj SCI_STYLESETBACK sel rgb
	method style_set_eolfilled sel mode =
		sci_send_message_ibu obj SCI_STYLESETEOLFILLED sel mode
	method style_set_characterset sel i =
		sci_send_message_iiu obj SCI_STYLESETCHARACTERSET sel (int_of_char_set i)
	method style_set_case sel i = sci_send_message_iAu obj SCI_STYLESETCASE sel i
	method style_set_visible sel mode =
		sci_send_message_ibu obj SCI_STYLESETVISIBLE sel mode
	method style_set_changeable sel mode =
		sci_send_message_ibu obj SCI_STYLESETCHANGEABLE sel mode
	method style_set_hotspot sel mode =
		sci_send_message_ibu obj SCI_STYLESETHOTSPOT sel mode

	(* Caret, selection, and hotspot styles *)
	method set_sel_fore mode rgb = sci_send_message_biu obj SCI_SETSELFORE mode rgb
	method set_sel_back mode rgb = sci_send_message_biu obj SCI_SETSELBACK mode rgb
	method set_sel_alpha alpha = sci_send_message_iiu obj SCI_SETSELALPHA alpha 0
	method get_sel_alpha = sci_send_message_iii obj SCI_GETSELALPHA 0 0
	method set_caret_fore rgb = sci_send_message_iiu obj SCI_SETCARETFORE rgb 0
	method get_caret_fore = sci_send_message_iii obj SCI_GETCARETFORE 0 0
	method set_caret_line_visible mode =
		sci_send_message_biu obj SCI_SETCARETLINEVISIBLE mode 0
	method get_caret_line_visible = sci_send_message_iib obj SCI_GETCARETLINEVISIBLE 0 0
	method set_caret_line_back rgb = sci_send_message_iiu obj SCI_SETCARETLINEBACK rgb 0
	method get_caret_line_back = sci_send_message_iii obj SCI_GETCARETLINEBACK 0 0
	method set_caret_line_back_alpha alpha =
		sci_send_message_iiu obj SCI_SETCARETLINEBACKALPHA alpha 0
	method get_caret_line_back_alpha =
		sci_send_message_iii obj SCI_GETCARETLINEBACKALPHA 0 0
	method set_caret_period len = sci_send_message_iiu obj SCI_SETCARETPERIOD len 0
	method get_caret_period = sci_send_message_iii obj SCI_GETCARETPERIOD 0 0
	method set_caret_width x = sci_send_message_iiu obj SCI_SETCARETWIDTH x 0
	method get_caret_width = sci_send_message_iii obj SCI_GETCARETWIDTH 0 0
	method set_hotspot_active_fore mode rgb =
		sci_send_message_biu obj SCI_SETHOTSPOTACTIVEFORE mode rgb
	method set_hotspot_active_back mode rgb =
		sci_send_message_biu obj SCI_SETHOTSPOTACTIVEBACK mode rgb
	method set_hotspot_active_underline mode =
		sci_send_message_biu obj SCI_SETHOTSPOTACTIVEUNDERLINE mode 0
	method set_control_char_symbol mode =
		sci_send_message_biu obj SCI_SETCONTROLCHARSYMBOL mode 0
	method get_control_char_symbol =
		sci_send_message_iic obj SCI_GETCONTROLCHARSYMBOL 0 0

	(* Margins *)
	method set_margin_type_n sel i = sci_send_message_iMu obj SCI_SETMARGINTYPEN sel i
	method get_margin_type_n = sci_send_message_iiM obj SCI_GETMARGINTYPEN 0 0
	method set_margin_width_n sel x = sci_send_message_iiu obj SCI_SETMARGINWIDTHN sel x
	method get_margin_width_n = sci_send_message_iii obj SCI_GETMARGINWIDTHN 0 0
	method set_margin_mask_n sel mask =
		sci_send_message_i3u obj SCI_SETMARGINMASKN sel mask
	method get_margin_mask_n = sci_send_message_ii3 obj SCI_GETMARGINMASKN 0 0
	method set_margin_sensitive_n sel mode =
		sci_send_message_ibu obj SCI_SETMARGINSENSITIVEN sel mode
	method get_margin_sensitive_n = sci_send_message_iib obj SCI_GETMARGINSENSITIVEN 0 0
	method set_margin_left x = sci_send_message_iiu obj SCI_SETMARGINLEFT 0 x
	method get_margin_left = sci_send_message_iii obj SCI_GETMARGINLEFT 0 0
	method set_margin_right x = sci_send_message_iiu obj SCI_SETMARGINRIGHT 0 x
	method get_margin_right = sci_send_message_iii obj SCI_GETMARGINRIGHT 0 0
	method set_fold_margin_colour mode rgb =
		sci_send_message_biu obj SCI_SETFOLDMARGINCOLOUR mode rgb
	method set_fold_margin_hi_colour mode rgb =
		sci_send_message_biu obj SCI_SETFOLDMARGINHICOLOUR mode rgb

	(* Other settings *)
	method set_use_palette mode = sci_send_message_biu obj SCI_SETUSEPALETTE mode 0
	method get_use_palette = sci_send_message_iib obj SCI_GETUSEPALETTE 0 0
	method set_buffered_draw mode = sci_send_message_biu obj SCI_SETBUFFEREDDRAW mode 0
	method get_buffered_draw = sci_send_message_iib obj SCI_GETBUFFEREDDRAW 0 0
	method set_two_phase_draw mode = sci_send_message_biu obj SCI_SETTWOPHASEDRAW mode 0
	method get_two_phase_draw = sci_send_message_iib obj SCI_GETTWOPHASEDRAW 0 0
	method set_code_page cp = sci_send_message_iiu obj SCI_SETCODEPAGE cp 0
	method get_code_page = sci_send_message_iii obj SCI_GETCODEPAGE 0 0
	method set_word_chars buf = sci_send_message_isu obj SCI_SETWORDCHARS 0 buf
	method set_whitespace_chars buf =
		sci_send_message_isu obj SCI_SETWHITESPACECHARS 0 buf
	method set_chars_default () = sci_send_message_iiu obj SCI_SETCHARSDEFAULT 0 0
	method grab_focus () = sci_send_message_iiu obj SCI_GRABFOCUS 0 0
	method set_focus mode = sci_send_message_biu obj SCI_SETFOCUS mode 0
	method get_focus = sci_send_message_iib obj SCI_GETFOCUS 0 0

	(* Brace highlighting *)
	method brace_highlight pos1 pos2 =
		sci_send_message_iiu obj SCI_BRACEHIGHLIGHT pos1 pos2
	method brace_badlight pos = sci_send_message_iiu obj SCI_BRACEBADLIGHT pos 0
	method brace_match pos mode = sci_send_message_iii obj SCI_BRACEMATCH pos mode

	(* Tabs and Indentation Guides *)
	method set_tab_width col = sci_send_message_iiu obj SCI_SETTABWIDTH col 0
	method get_tab_width = sci_send_message_iii obj SCI_GETTABWIDTH 0 0
	method set_use_tabs mode = sci_send_message_biu obj SCI_SETUSETABS mode 0
	method get_use_tabs = sci_send_message_iib obj SCI_GETUSETABS 0 0
	method set_indent col = sci_send_message_iiu obj SCI_SETINDENT col 0
	method get_indent = sci_send_message_iii obj SCI_GETINDENT 0 0
	method set_tab_indents mode = sci_send_message_biu obj SCI_SETTABINDENTS mode 0
	method get_tab_indents = sci_send_message_iib obj SCI_GETTABINDENTS 0 0
	method set_backspace_unindents mode =
		sci_send_message_biu obj SCI_SETBACKSPACEUNINDENTS mode 0
	method get_backspace_unindents =
		sci_send_message_iib obj SCI_GETBACKSPACEUNINDENTS 0 0
	method set_line_indentation line col =
		sci_send_message_iiu obj SCI_SETLINEINDENTATION line col
	method get_line_indentation line =
		sci_send_message_iii obj SCI_GETLINEINDENTATION line 0
	method get_line_indentation_position line =
		sci_send_message_iii obj SCI_GETLINEINDENTPOSITION line 0
	method set_indentation_guides mode =
		sci_send_message_biu obj SCI_SETINDENTATIONGUIDES mode 0
	method get_indentation_guides =
		sci_send_message_iib obj SCI_GETINDENTATIONGUIDES 0 0
	method set_highlight_guide col =
		sci_send_message_iiu obj SCI_SETHIGHLIGHTGUIDE col 0
	method get_highlight_guide = sci_send_message_iii obj SCI_GETHIGHLIGHTGUIDE 0 0

	(* Markers *)
	method marker_define sel m =
		sci_send_message_iiu obj SCI_MARKERDEFINE sel (int_of_marker_symbol m)
	method marker_define_pixmap sel buf =
		sci_send_message_iiu obj SCI_MARKERDEFINEPIXMAP sel buf
	method marker_set_fore sel rgb = sci_send_message_iiu obj SCI_MARKERSETFORE sel rgb
	method marker_set_back sel rgb = sci_send_message_iiu obj SCI_MARKERSETBACK sel rgb
	method marker_set_alpha sel alpha =
		sci_send_message_iiu obj SCI_MARKERSETALPHA sel alpha
	method marker_add line sel = sci_send_message_iii obj SCI_MARKERADD line sel
	method marker_add_set line mask =
		sci_send_message_i3u obj SCI_MARKERADDSET line mask
	method marker_delete line sel = sci_send_message_iiu obj SCI_MARKERDELETE line sel
	method marker_delete_all sel = sci_send_message_iiu obj SCI_MARKERDELETEALL sel 0
	method marker_get line = sci_send_message_ii3 obj SCI_MARKERGET line 0
	method marker_next line mask = sci_send_message_i3i obj SCI_MARKERNEXT line mask
	method marker_previous line mask =
		sci_send_message_i3i obj SCI_MARKERPREVIOUS line mask
	method marker_line_from_handle h =
		sci_send_message_iii obj SCI_MARKERLINEFROMHANDLE h 0
	method marker_delete_handle h = sci_send_message_iiu obj SCI_MARKERDELETEHANDLE h 0

	(* Indicators *)
	method indic_set_style sel mode =
		sci_send_message_iIu obj SCI_INDICSETSTYLE sel mode
	method indic_get_style sel = sci_send_message_iiI obj SCI_INDICGETSTYLE sel 0
	method indic_set_fore sel rgb = sci_send_message_iiu obj SCI_INDICSETFORE sel rgb
	method indic_get_fore sel = sci_send_message_iii obj SCI_INDICGETFORE sel 0

	(* Autocompletion *)
	method autoc_show len buf = sci_send_message_iiu obj SCI_AUTOCSHOW len buf
	method autoc_cancel () = sci_send_message_iiu obj SCI_AUTOCCANCEL 0 0
	method autoc_active = sci_send_message_iii obj SCI_AUTOCACTIVE 0 0
	method autoc_pos_start = sci_send_message_iii obj SCI_AUTOCPOSSTART 0 0
	method autoc_complete () = sci_send_message_iiu obj SCI_AUTOCCOMPLETE 0 0
	method autoc_stop_s buf = sci_send_message_iiu obj SCI_AUTOCSTOPS 0 buf
	method autoc_set_separator c = sci_send_message_iiu obj SCI_AUTOCSETSEPARATOR c 0
	method autoc_get_separator = sci_send_message_iic obj SCI_AUTOCGETSEPARATOR 0 0
	method autoc_select buf = sci_send_message_iiu obj SCI_AUTOCSELECT 0 buf
	method autoc_get_current = sci_send_message_iii obj SCI_AUTOCGETCURRENT 0 0
	method autoc_set_cancel_at_start mode =
		sci_send_message_biu obj SCI_AUTOCSETCANCELATSTART mode 0
	method autoc_get_cancel_at_start =
		sci_send_message_iib obj SCI_AUTOCGETCANCELATSTART 0 0
	method autoc_set_fillups buf = sci_send_message_iiu obj SCI_AUTOCSETFILLUPS 0 buf
	method autoc_set_choose_single mode =
		sci_send_message_biu obj SCI_AUTOCSETCHOOSESINGLE mode 0
	method autoc_get_choose_single =
		sci_send_message_iib obj SCI_AUTOCGETCHOOSESINGLE 0 0
	method autoc_set_ignore_case mode =
		sci_send_message_biu obj SCI_AUTOCSETIGNORECASE mode 0
	method autoc_get_ignore_case = sci_send_message_iib obj SCI_AUTOCGETIGNORECASE 0 0
	method autoc_set_auto_hide mode =
		sci_send_message_biu obj SCI_AUTOCSETAUTOHIDE mode 0
	method autoc_get_auto_hide = sci_send_message_iib obj SCI_AUTOCGETAUTOHIDE 0 0
	method autoc_set_drop_rest_of_word mode =
		sci_send_message_biu obj SCI_AUTOCSETDROPRESTOFWORD mode 0
	method autoc_get_drop_rest_of_word =
		sci_send_message_iib obj SCI_AUTOCGETDROPRESTOFWORD 0 0
	method register_image sel buf = sci_send_message_isu obj SCI_REGISTERIMAGE sel buf
	method clear_registered_images () =
		sci_send_message_iiu obj SCI_CLEARREGISTEREDIMAGES 0 0
	method autoc_set_type_separator c =
		sci_send_message_iiu obj SCI_AUTOCSETTYPESEPARATOR c 0
	method autoc_get_type_separator =
		sci_send_message_iic obj SCI_AUTOCGETTYPESEPARATOR 0 0

	(* User lists *)
	method user_list_show sel buf = sci_send_message_isu obj SCI_USERLISTSHOW sel buf

	(* Call tips *)
	method call_tip_show pos buf = sci_send_message_isu obj SCI_CALLTIPSHOW pos buf
	method call_tip_cancel () = sci_send_message_iiu obj SCI_CALLTIPCANCEL 0 0
	method call_tip_active = sci_send_message_iii obj SCI_CALLTIPACTIVE 0 0
	method call_tip_pos_start = sci_send_message_iii obj SCI_CALLTIPPOSSTART 0 0
	method call_tip_set_hlt pos1 pos2 =
		sci_send_message_iiu obj SCI_CALLTIPSETHLT pos1 pos2
	method call_tip_set_back rgb = sci_send_message_iiu obj SCI_CALLTIPSETBACK rgb 0
	method call_tip_set_fore rgb = sci_send_message_iiu obj SCI_CALLTIPSETFORE rgb 0
	method call_tip_set_fore_hlt rgb =
		sci_send_message_iiu obj SCI_CALLTIPSETFOREHLT rgb 0

	(* Keyboard commands *)
	(* N.B. - this "meta" method supplements the standard Scintilla API *)
	method send_key_cmd cmd = sci_send_message_key obj cmd

	(* Key bindings *)
	(* N.B. - restricted to key_cmd values ONLY - NO sci_cmd values *)
	method assign_cmd_key key cmd =
		sci_send_message_iiu obj SCI_ASSIGNCMDKEY
			(int_of_key_definition key) (int_of_key_cmd cmd)
	method clear_cmd_key key =
		sci_send_message_iiu obj SCI_CLEARCMDKEY (int_of_key_definition key) 0
	method clear_all_cmd_keys () = sci_send_message_iiu obj SCI_CLEARALLCMDKEYS 0 0
	method null = sci_send_message_iiu obj SCI_NULL 0 0

	(* Popup edit menu *)
	method use_popup mode = sci_send_message_biu obj SCI_USEPOPUP mode 0

	(* Macro recording *)
	method start_record () = sci_send_message_iiu obj SCI_STARTRECORD 0 0
	method stop_record () = sci_send_message_iiu obj SCI_STOPRECORD 0 0

	(* Printing - Windows ONLY *)

	(* Direct access - Windows ONLY *)

	(* Multiple views *)
	method get_doc_pointer = sci_send_message_iid obj SCI_GETDOCPOINTER 0 0
	method set_doc_pointer doc = sci_send_message_idu obj SCI_SETDOCPOINTER 0 doc
	method create_document () = sci_send_message_iid obj SCI_CREATEDOCUMENT 0 0
	method add_ref_document doc = sci_send_message_idu obj SCI_ADDREFDOCUMENT 0 doc
	method release_document doc = sci_send_message_idu obj SCI_RELEASEDOCUMENT 0 doc

	(* Folding *)
	method visible_from_doc_line line =
		sci_send_message_iii obj SCI_VISIBLEFROMDOCLINE line 0
	method doc_line_from_visible line =
		sci_send_message_iii obj SCI_DOCLINEFROMVISIBLE line 0
	method show_lines line1 line2 = sci_send_message_iiu obj SCI_SHOWLINES line1 line2
	method hide_lines line1 line2 = sci_send_message_iiu obj SCI_HIDELINES line1 line2
	method get_line_visible line = sci_send_message_iii obj SCI_GETLINEVISIBLE line 0
	method set_fold_level line i = sci_send_message_iiu obj SCI_SETFOLDLEVEL line i
	method get_fold_level line = sci_send_message_iii obj SCI_GETFOLDLEVEL line 0
	method set_fold_flags mode = sci_send_message_iiu obj SCI_SETFOLDFLAGS mode 0
	method get_last_child line i = sci_send_message_iii obj SCI_GETLASTCHILD line i
	method get_fold_parent line = sci_send_message_iii obj SCI_GETFOLDPARENT line 0
	method toggle_fold line = sci_send_message_iiu obj SCI_TOGGLEFOLD line 0
	method set_fold_expanded line mode =
		sci_send_message_ibu obj SCI_SETFOLDEXPANDED line mode
	method get_fold_expanded line = sci_send_message_iib obj SCI_GETFOLDEXPANDED line 0
	method ensure_visible line = sci_send_message_iiu obj SCI_ENSUREVISIBLE line 0
	method ensure_visible_enforce_policy line =
		sci_send_message_iiu obj SCI_ENSUREVISIBLEENFORCEPOLICY line 0

	(* Line wrapping *)
	method set_wrap_mode mode = sci_send_message_Uiu obj SCI_SETWRAPMODE mode 0
	method get_wrap_mode = sci_send_message_iiU obj SCI_GETWRAPMODE 0 0
	method set_wrap_visual_flags ?(flags=[]) () =
		sci_send_message_iiu obj SCI_SETWRAPVISUALFLAGS (int_of_visual_flag_list flags) 0
	method get_wrap_visual_flags =
		explode_all_exp visual_flag_of_int
			(sci_send_message_iii obj SCI_GETWRAPVISUALFLAGS 0 0)
	method set_wrap_visual_flags_location ?(flags=[]) () =
		sci_send_message_iiu obj SCI_SETWRAPVISUALFLAGSLOCATION
			(int_of_visual_flag_loc_list flags) 0
	method get_wrap_visual_flags_location =
		explode_all_exp visual_flag_loc_of_int
			(sci_send_message_iii obj SCI_GETWRAPVISUALFLAGSLOCATION 0 0)
	method set_wrap_start_indent col =
		sci_send_message_iiu obj SCI_SETWRAPSTARTINDENT col 0
	method get_wrap_start_indent = sci_send_message_iii obj SCI_GETWRAPSTARTINDENT 0 0
	method set_layout_cache mode = sci_send_message_Ciu obj SCI_SETLAYOUTCACHE mode 0
	method get_layout_cache = sci_send_message_iiC obj SCI_GETLAYOUTCACHE 0 0
	method lines_split x = sci_send_message_iiu obj SCI_LINESSPLIT x 0
	method lines_join () = sci_send_message_iiu obj SCI_LINESJOIN 0 0

	(* Zooming *)
	method zoom_in () = sci_send_message_iiu obj SCI_ZOOMIN 0 0
	method zoom_out () = sci_send_message_iiu obj SCI_ZOOMOUT 0 0
	method set_zoom y = sci_send_message_iiu obj SCI_SETZOOM y 0
	method get_zoom = sci_send_message_iii obj SCI_GETZOOM 0 0

	(* Long lines *)
	method set_edge_mode mode = sci_send_message_Eiu obj SCI_SETEDGEMODE mode 0
	method get_edge_mode = sci_send_message_iiE obj SCI_GETEDGEMODE 0 0
	method set_edge_column col = sci_send_message_iiu obj SCI_SETEDGECOLUMN col 0
	method get_edge_column = sci_send_message_iii obj SCI_GETEDGECOLUMN 0 0
	method set_edge_colour rgb = sci_send_message_iiu obj SCI_SETEDGECOLOUR rgb 0
	method get_edge_colour = sci_send_message_iii obj SCI_GETEDGECOLOUR 0 0

	(* Lexer *)
	method set_lexer i = sci_send_message_iiu obj SCI_SETLEXER i 0
	method get_lexer = sci_send_message_iii obj SCI_GETLEXER 0 0
	method set_lexer_language buf = sci_send_message_isu obj SCI_SETLEXERLANGUAGE 0 buf
	method load_lexer_library buf = sci_send_message_isu obj SCI_LOADLEXERLIBRARY 0 buf
	method colourise pos1 pos2 = sci_send_message_iiu obj SCI_COLOURISE pos1 pos2
	method set_property key buf = sci_send_message_ssu obj SCI_SETPROPERTY key buf
	method get_property key =
		let n = sci_send_message_sii obj SCI_GETPROPERTY key 0 in
		if n <= 0 then ""
		else
			let buf = String.create n in
			sci_send_message_ssi obj SCI_GETPROPERTY key buf; buf
	method get_property_expanded key =
		let n = sci_send_message_sii obj SCI_GETPROPERTYEXPANDED key 0 in
		if n <= 0 then ""
		else
			let buf = String.create n in
			sci_send_message_ssi obj SCI_GETPROPERTYEXPANDED key buf; buf
	method get_property_int ?(default=0) key =
		sci_send_message_sii obj SCI_GETPROPERTYINT key default
	method set_keywords sel buf = sci_send_message_isu obj SCI_SETKEYWORDS sel buf
	method get_style_bits_needed = sci_send_message_iii obj SCI_GETSTYLEBITSNEEDED 0 0

	(* Notifications *)
	method set_mod_event_mask ?(flags=[]) () =
		sci_send_message_iiu obj SCI_SETMODEVENTMASK (int_of_modification_type_list flags) 0
	method get_mod_event_mask =
		explode_all_exp modification_type_of_int
			(sci_send_message_iii obj SCI_GETMODEVENTMASK 0 0)
	method set_mouse_dwell_time t = sci_send_message_iiu obj SCI_SETMOUSEDWELLTIME t 0
	method get_mouse_dwell_time = sci_send_message_iii obj SCI_GETMOUSEDWELLTIME 0 0
end

(* get [GObject] type info *)
let get_type = sci_get_type ()

(* create a new editor object instance *)
let editor ?packing () =
	let self = new editor (sci_new ()) in
	may packing ~f:(fun f -> (f self#coerce));
	self

(* the mysterious "release resources" - use when all instances are done *)
let release_resources () = sci_release_resources ()
